import os
from enum import Enum


class Modes(Enum):
    NATIVE_TEXT = "native_text"
    LOW_COST = "low_cost"
    HIGH_QUALITY = "high_quality"
    FORM = "form"


class OutputModes(Enum):
    LAYOUT_PRESERVING = "layout_preserving"
    TEXT = "text"


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"


class WhispererHeader:
    UNSTRACT_KEY = "unstract-key"


class WhispererEndpoint:
    """Endpoints available at LLMWhisperer service."""

    TEST_CONNECTION = "test-connection"
    WHISPER = "whisper"
    STATUS = "whisper-status"
    RETRIEVE = "whisper-retrieve"
    HIGHLIGHTS = "highlights"


class WhispererEnv:
    """Env variables for LLMWhisperer.

    Can be used to alter behaviour at runtime.

    Attributes:
        WAIT_TIMEOUT: Timeout for the extraction in seconds. Defaults to 300s
        LOG_LEVEL: Logging level for the client library. Defaults to INFO
    """

    WAIT_TIMEOUT = "ADAPTER_LLMW_WAIT_TIMEOUT"
    LOG_LEVEL = "LOG_LEVEL"


class WhispererConfig:
    """Dictionary keys used to configure LLMWhisperer service."""

    URL = "url"
    MODE = "mode"
    OUTPUT_MODE = "output_mode"
    UNSTRACT_KEY = "unstract_key"
    MEDIAN_FILTER_SIZE = "median_filter_size"
    GAUSSIAN_BLUR_RADIUS = "gaussian_blur_radius"
    LINE_SPLITTER_TOLERANCE = "line_splitter_tolerance"
    LINE_SPLITTER_STRATEGY = "line_spitter_strategy"
    HORIZONTAL_STRETCH_FACTOR = "horizontal_stretch_factor"
    PAGES_TO_EXTRACT = "pages_to_extract"
    MARK_VERTICAL_LINES = "mark_vertical_lines"
    MARK_HORIZONTAL_LINES = "mark_horizontal_lines"
    PAGE_SEPARATOR = "page_seperator"
    URL_IN_POST = "url_in_post"
    TAG = "tag"
    USE_WEBHOOK = "use_webhook"
    WEBHOOK_METADATA = "webhook_metadata"
    TEXT_ONLY = "text_only"
    WAIT_TIMEOUT = "wait_timeout"
    WAIT_FOR_COMPLETION = "wait_for_completion"
    LOGGING_LEVEL = "logging_level"
    ADD_LINE_NOS = "add_line_nos"
    EXTRACT_ALL_LINES = "extract_all_lines"
    LINES = "lines"


class WhisperStatus:
    """Values returned / used by /whisper-status endpoint."""

    PROCESSING = "processing"
    PROCESSED = "processed"
    DELIVERED = "delivered"
    UNKNOWN = "unknown"
    # Used for async processing
    WHISPER_HASH = "whisper_hash"
    STATUS = "status"


class WhispererDefaults:
    """Defaults meant for LLMWhisperer."""

    MEDIAN_FILTER_SIZE = 0
    GAUSSIAN_BLUR_RADIUS = 0.0
    FORCE_TEXT_PROCESSING = False
    LINE_SPLITTER_TOLERANCE = 0.75
    LINE_SPLITTER_STRATEGY = "left-priority"
    HORIZONTAL_STRETCH_FACTOR = 1.0
    PAGES_TO_EXTRACT = ""
    PAGE_SEPARATOR = "<<<"
    MARK_VERTICAL_LINES = False
    MARK_HORIZONTAL_LINES = False
    URL_IN_POST = False
    TAG = "default"
    TEXT_ONLY = False
    WAIT_TIMEOUT = int(os.getenv(WhispererEnv.WAIT_TIMEOUT, 900))
    WAIT_FOR_COMPLETION = True
    LOGGING_LEVEL = os.getenv(WhispererEnv.LOG_LEVEL, "INFO")



import os
from typing import Any

from llama_index.core.llms import LLM
from llama_index.llms.azure_openai import AzureOpenAI
from llama_index.llms.openai.utils import O1_MODELS

from unstract.sdk.adapters.exceptions import AdapterError
from unstract.sdk.adapters.llm.constants import LLMKeys
from unstract.sdk.adapters.llm.llm_adapter import LLMAdapter


class Constants:
    MODEL = "model"
    DEPLOYMENT_NAME = "deployment_name"
    API_KEY = "api_key"
    API_VERSION = "api_version"
    MAX_RETRIES = "max_retries"
    MAX_TOKENS = "max_tokens"
    AZURE_ENDPONT = "azure_endpoint"
    API_TYPE = "azure"
    TIMEOUT = "timeout"
    DEFAULT_MODEL = "gpt-35-turbo"
    ENABLE_REASONING = "enable_reasoning"
    REASONING_EFFORT = "reasoning_effort"


class AzureOpenAILLM(LLMAdapter):
    def __init__(self, settings: dict[str, Any]):
        super().__init__("AzureOpenAI")
        self.config = settings

    SCHEMA_PATH = f"{os.path.dirname(__file__)}/static/json_schema.json"

    @staticmethod
    def get_id() -> str:
        return "azureopenai|592d84b9-fe03-4102-a17e-6b391f32850b"

    @staticmethod
    def get_name() -> str:
        return "AzureOpenAI"

    @staticmethod
    def get_description() -> str:
        return "AzureOpenAI LLM"

    @staticmethod
    def get_provider() -> str:
        return "azure"

    @staticmethod
    def get_icon() -> str:
        return "/icons/adapter-icons/AzureopenAI.png"

    def get_llm_instance(self) -> LLM:
        max_retries = int(
            self.config.get(Constants.MAX_RETRIES, LLMKeys.DEFAULT_MAX_RETRIES)
        )
        max_tokens = self.config.get(Constants.MAX_TOKENS)
        max_tokens = int(max_tokens) if max_tokens else None
        enable_reasoning = self.config.get(Constants.ENABLE_REASONING)
        model = self.config.get(Constants.MODEL, Constants.DEFAULT_MODEL)

        llm_kwargs = {
            "model": model,
            "deployment_name": str(self.config.get(Constants.DEPLOYMENT_NAME)),
            "api_key": str(self.config.get(Constants.API_KEY)),
            "api_version": str(self.config.get(Constants.API_VERSION)),
            "azure_endpoint": str(self.config.get(Constants.AZURE_ENDPONT)),
            "api_type": Constants.API_TYPE,
            "temperature": 0,
            "timeout": float(self.config.get(Constants.TIMEOUT, LLMKeys.DEFAULT_TIMEOUT)),
            "max_retries": max_retries,
        }

        if enable_reasoning:
            llm_kwargs["reasoning_effort"] = self.config.get(
                    Constants.REASONING_EFFORT
                )

        if model not in O1_MODELS:
            llm_kwargs["max_completion_tokens"] = max_tokens
        else:
            llm_kwargs["max_tokens"] = max_tokens

        try:
            llm: LLM = AzureOpenAI(**llm_kwargs)
            return llm
        except Exception as e:
            raise AdapterError(str(e))



from dataclasses import dataclass


@dataclass
class WhispererRequestParams:
    """DTO for LLM Whisperer API request parameters.

    Args:
        tag (Optional[Union[str, List[str]]]): Tag value. Can be initialized with List[str] or str.
             Will be converted to str or None after initialization.
        enable_highlight (bool): Whether to enable highlighting. Defaults to False.
    """

    # TODO: Extend this DTO to include all Whisperer API parameters
    tag: str | list[str] | None = None
    enable_highlight: bool = False

    def __post_init__(self) -> None:
        # TODO: Allow list of tags once it's supported in LLMW v2
        if isinstance(self.tag, list):
            self.tag = self.tag[0] if self.tag else None



import json

from unstract.sdk.adapter import ToolAdapter
from unstract.sdk.file_storage import FileStorage, FileStorageProvider
from unstract.sdk.tool.base import BaseTool
from unstract.sdk.utils import ToolUtils


class IndexingUtils:
    @staticmethod
    def generate_index_key(
        vector_db: str,
        embedding: str,
        x2text: str,
        chunk_size: str,
        chunk_overlap: str,
        tool: BaseTool,
        file_path: str | None = None,
        file_hash: str | None = None,
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> str:
        """Generates a unique index key based on the provided configuration,
        file information, instance identifiers, and processing options.

        Args:
            fs (FileStorage, optional): File storage for remote storage.

        Returns:
            str: A unique index key used for indexing the document.
        """
        if not file_path and not file_hash:
            raise ValueError("One of `file_path` or `file_hash` need to be provided")

        if not file_hash:
            file_hash = fs.get_hash_from_file(path=file_path)

        # Whole adapter config is used currently even though it contains some keys
        # which might not be relevant to indexing. This is easier for now than
        # marking certain keys of the adapter config as necessary.
        index_key = {
            "file_hash": file_hash,
            "vector_db_config": ToolAdapter.get_adapter_config(tool, vector_db),
            "embedding_config": ToolAdapter.get_adapter_config(tool, embedding),
            "x2text_config": ToolAdapter.get_adapter_config(tool, x2text),
            # Typed and hashed as strings since the final hash is persisted
            # and this is required to be backward compatible
            "chunk_size": str(chunk_size),
            "chunk_overlap": str(chunk_overlap),
        }
        # JSON keys are sorted to ensure that the same key gets hashed even in
        # case where the fields are reordered.
        hashed_index_key = ToolUtils.hash_str(json.dumps(index_key, sort_keys=True))
        return hashed_index_key



import functools
import logging
from collections.abc import Callable
from typing import Any, ParamSpec, TypeVar

import requests
from deprecated import deprecated
from requests import ConnectionError, RequestException, Response
from unstract.sdk.constants import (
    MimeType,
    RequestHeader,
    ToolEnv,
)
from unstract.sdk.helper import SdkHelper
from unstract.sdk.platform import PlatformHelper
from unstract.sdk.tool.base import BaseTool
from unstract.sdk.utils.common_utils import log_elapsed

logger = logging.getLogger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


def handle_service_exceptions(context: str) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to handle exceptions in PromptTool service calls.

    Args:
        context (str): Context string describing where the error occurred
    Returns:
        Callable: Decorated function that handles service exceptions
    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            try:
                return func(*args, **kwargs)
            except ConnectionError as e:
                msg = f"Error while {context}. Unable to connect to prompt service."
                logger.error(f"{msg}\n{e}")
                args[0].tool.stream_error_and_exit(msg, e)
            except RequestException as e:
                error_message = str(e)
                response = getattr(e, "response", None)
                if response is not None:
                    if (
                        MimeType.JSON in response.headers.get("Content-Type", "").lower()
                        and "error" in response.json()
                    ):
                        error_message = response.json()["error"]
                    elif response.text:
                        error_message = response.text
                msg = f"Error while {context}. {error_message}"
                args[0].tool.stream_error_and_exit(msg, e)

        return wrapper

    return decorator


class PromptTool:
    """Class to handle prompt service methods for Unstract Tools."""

    def __init__(
        self,
        tool: BaseTool,
        prompt_host: str,
        prompt_port: str,
        is_public_call: bool = False,
        request_id: str | None = None,
    ) -> None:
        """Class to interact with prompt-service.

        Args:
            tool (AbstractTool): Instance of AbstractTool
            prompt_host (str): Host of platform service
            prompt_port (str): Port of platform service
            is_public_call (bool): Whether the call is public. Defaults to False
        """
        self.tool = tool
        self.base_url = SdkHelper.get_platform_base_url(prompt_host, prompt_port)
        self.is_public_call = is_public_call
        self.request_id = request_id
        if not is_public_call:
            self.bearer_token = tool.get_env_or_die(ToolEnv.PLATFORM_API_KEY)

    @log_elapsed(operation="ANSWER_PROMPTS")
    @handle_service_exceptions("answering prompt(s)")
    def answer_prompt(
        self,
        payload: dict[str, Any],
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        url_path = "answer-prompt"
        if self.is_public_call:
            url_path = "answer-prompt-public"
        return self._call_service(
            url_path=url_path, payload=payload, params=params, headers=headers
        )

    @log_elapsed(operation="INDEX")
    @handle_service_exceptions("indexing")
    def index(
        self,
        payload: dict[str, Any],
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
    ) -> str:
        url_path = "index"
        if self.is_public_call:
            url_path = "index-public"
        prompt_service_response = self._call_service(
            url_path=url_path,
            payload=payload,
            params=params,
            headers=headers,
        )
        return prompt_service_response.get("doc_id")

    @log_elapsed(operation="EXTRACT")
    @handle_service_exceptions("extracting")
    def extract(
        self,
        payload: dict[str, Any],
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        url_path = "extract"
        if self.is_public_call:
            url_path = "extract-public"
        prompt_service_response = self._call_service(
            url_path=url_path,
            payload=payload,
            params=params,
            headers=headers,
        )
        return prompt_service_response.get("extracted_text")

    @log_elapsed(operation="SINGLE_PASS_EXTRACTION")
    @handle_service_exceptions("single pass extraction")
    def single_pass_extraction(
        self,
        payload: dict[str, Any],
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        return self._call_service(
            url_path="single-pass-extraction",
            payload=payload,
            params=params,
            headers=headers,
        )

    @log_elapsed(operation="SUMMARIZATION")
    @handle_service_exceptions("summarizing")
    def summarize(
        self,
        payload: dict[str, Any],
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        return self._call_service(
            url_path="summarize",
            payload=payload,
            params=params,
            headers=headers,
        )

    def _get_headers(self, headers: dict[str, str] | None = None) -> dict[str, str]:
        """Get default headers for requests.

        Returns:
            dict[str, str]: Default headers including request ID and authorization
        """
        request_headers = {RequestHeader.REQUEST_ID: self.request_id}
        if self.is_public_call:
            return request_headers
        request_headers.update(
            {RequestHeader.AUTHORIZATION: f"Bearer {self.bearer_token}"}
        )

        if headers:
            request_headers.update(headers)
        return request_headers

    def _call_service(
        self,
        url_path: str,
        payload: dict[str, Any] | None = None,
        params: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        method: str = "POST",
    ) -> dict[str, Any]:
        """Communicates to prompt service to fetch response for the prompt.

        Only POST calls are made to prompt-service though functionality exists.

        Args:
            url_path (str): URL path to the service endpoint
            payload (dict, optional): Payload to send in the request body
            params (dict, optional): Query parameters to include in the request
            headers (dict, optional): Headers to include in the request
            method (str): HTTP method to use for the request (GET or POST)

        Returns:
            dict: Response from the prompt service
        """
        url: str = f"{self.base_url}/{url_path}"
        req_headers = self._get_headers(headers)
        response: Response = Response()
        if method.upper() == "POST":
            response = requests.post(
                url=url, json=payload, params=params, headers=req_headers
            )
        elif method.upper() == "GET":
            response = requests.get(url=url, params=params, headers=req_headers)
        else:
            raise ValueError(f"Unsupported HTTP method: {method}")

        response.raise_for_status()
        return response.json()

    @staticmethod
    @deprecated(
        version="v0.71.0", reason="Use `PlatformHelper.get_prompt_studio_tool` instead"
    )
    def get_exported_tool(
        tool: BaseTool, prompt_registry_id: str
    ) -> dict[str, Any] | None:
        """Get exported custom tool by the help of unstract DB tool.

        Args:
            prompt_registry_id (str): ID of the prompt_registry_id
            tool (AbstractTool): Instance of AbstractTool
        Required env variables:
            PLATFORM_HOST: Host of platform service
            PLATFORM_PORT: Port of platform service
        """
        platform_helper: PlatformHelper = PlatformHelper(
            tool=tool,
            platform_port=tool.get_env_or_die(ToolEnv.PLATFORM_PORT),
            platform_host=tool.get_env_or_die(ToolEnv.PLATFORM_HOST),
        )
        return platform_helper.get_prompt_studio_tool(
            prompt_registry_id=prompt_registry_id
        )



import json
import logging
from datetime import datetime
from hashlib import sha256
from typing import Any

import filetype
import fsspec
import magic
import yaml
from unstract.sdk.exceptions import FileOperationError
from unstract.sdk.file_storage.constants import FileOperationParams, FileSeekPosition
from unstract.sdk.file_storage.helper import FileStorageHelper, skip_local_cache
from unstract.sdk.file_storage.interface import FileStorageInterface
from unstract.sdk.file_storage.provider import FileStorageProvider

logger = logging.getLogger(__name__)


class FileStorage(FileStorageInterface):
    # This class integrates fsspec library for file operations

    fs: fsspec  # fsspec file system handle
    provider: FileStorageProvider

    def __init__(self, provider: FileStorageProvider, **storage_config: dict[str, Any]):
        self.fs = FileStorageHelper.file_storage_init(provider, **storage_config)
        self.provider = provider

    @skip_local_cache
    def read(
        self,
        path: str,
        mode: str,
        encoding: str = FileOperationParams.DEFAULT_ENCODING,
        seek_position: int = 0,
        length: int = FileOperationParams.READ_ENTIRE_LENGTH,
    ) -> bytes | str:
        """Read the file pointed to by the file_handle.

        Args:
            path (str): Path to the file to be opened
            mode (str): Mode in which the file is to be opened. Usual options
                        include r, rb, w and wb
            encoding (str): Encoding type like utf-8 or utf-16
            seek_position (int): Position to start reading from
            length (int): Number of bytes to be read. Default is full
            file content.

        Returns:
            Union[bytes, str] - File contents in bytes/string based on the opened mode
        """
        with self.fs.open(path=path, mode=mode, encoding=encoding) as file_handle:
            if seek_position > 0:
                file_handle.seek(seek_position)
            return file_handle.read(length)

    def write(
        self,
        path: str,
        mode: str,
        encoding: str = FileOperationParams.DEFAULT_ENCODING,
        seek_position: int = 0,
        data: bytes | str = "",
    ) -> int:
        """Write data in the file pointed to by the file-handle.

        Args:
            path (str): Path to the file to be opened
            mode (str): Mode in whicg the file is to be opened. Usual options
                        include r, rb, w and wb
            encoding (str): Encoding type like utf-8 or utf-16
            seek_position (int): Position to start writing from
            data (Union[bytes, str]): Contents to be written

        Returns:
            int: Number of bytes that were successfully written to the file
        """
        try:
            with self.fs.open(path=path, mode=mode, encoding=encoding) as file_handle:
                return file_handle.write(data)
        except Exception as e:
            raise FileOperationError(str(e)) from e

    @skip_local_cache
    def seek(
        self,
        path: str,
        location: int = 0,
        position: FileSeekPosition = FileSeekPosition.START,
    ) -> int:
        """Place the file pointer to the mentioned location in the file
        relative to the position.

        Args:
            path (str): path of the file
            location (int): Nth byte position. To be understood in relation to
            the arg "position"
            position (FileSeekPosition): from start of file, current location
            or end of file

        Returns:
            int: file pointer location after seeking to the mentioned position
        """
        with self.fs.open(path=path, mode="rb") as file_handle:
            return file_handle.seek(location, position)

    def mkdir(self, path: str, create_parents: bool = True):
        """Create a directory.

        Args:
            path (str): Path of the directory to be created
            create_parents (bool): Specify if parent directories to be created
            if any of the nested directory does not exist
        """
        try:
            self.fs.mkdir(path=path, create_parents=create_parents)
        except FileExistsError:
            logger.debug(f"Path {path} already exists.")
        except Exception as e:
            raise FileOperationError(str(e)) from e

    @skip_local_cache
    def exists(self, path: str) -> bool:
        """Checks if a file/directory path exists.

        Args:
            path (str): File/directory path

        Returns:
            bool: If the file/directory  exists or not
        """
        try:
            return self.fs.exists(path)
        except Exception as e:
            raise FileOperationError(str(e)) from e

    @skip_local_cache
    def ls(self, path: str) -> list[str]:
        """List the directory path.

        Args:
            path (str): Directory path

        Returns:
            List[str]: List of files / directories under the path
        """
        return self.fs.ls(path)

    @skip_local_cache
    def rm(self, path: str, recursive: bool = True):
        """Removes a file or directory mentioned in path.

        Args:
            path (str): Path to the file / directory
            recursive (bool): Whether the files and folders nested
            under path are to be removed or not

        Returns:
            NA
        """
        return self.fs.rm(path=path, recursive=recursive)

    @skip_local_cache
    def cp(
        self,
        src: str,
        dest: str,
        recursive: bool = False,
        overwrite: bool = True,
    ):
        """Copies files from source(lpath) path to the destination(rpath) path.

        Args:
            src (str): Path to the source
            dest (str): Path to the destination
            recursive (bool): Copy recursively when set to True
            overwrite (bool): Overwrite existing path with same name

        Returns:
            NA
        """
        return self.fs.cp(src, dest, recursive=recursive, overwrite=overwrite)

    @skip_local_cache
    def size(self, path: str) -> int:
        """Get the size of the file specified in path.

        Args:
            path (str): Path to the file

        Returns:
            int: Size of the file in bytes
        """
        file_info = self.fs.info(path)
        return file_info["size"]

    @skip_local_cache
    def modification_time(self, path: str) -> datetime:
        """Get the last modification time of the file specified in path.

        Args:
            path (str): Path to the file

        Returns:
            datetime: Last modified time in datetime
        """
        file_info = self.fs.info(path)

        # Try different possible timestamp keys
        file_mtime = None
        for key in ["mtime", "LastModified", "last_modified"]:
            file_mtime = file_info.get(key)
            if file_mtime is not None:
                break

        if file_mtime is None:
            raise FileOperationError(
                f"Could not find modification time in file info: {file_info}"
            )

        if isinstance(file_mtime, datetime):
            return file_mtime

        return datetime.fromtimestamp(file_mtime)

    def mime_type(
        self,
        path: str,
        read_length: int = FileOperationParams.READ_ENTIRE_LENGTH,
    ) -> str:
        """Gets the file MIME type for an input file. Uses libmagic to perform
        the same.

        Args:
            path (str): Path of the input file
            read_length (int): Length(bytes) to be read from the file for in
            order to identify the mime type. Defaults to read the entire length.

        Returns:
            str: MIME type of the file
        """
        sample_contents = self.read(path=path, mode="rb", length=read_length)
        mime_type = magic.from_buffer(sample_contents, mime=True)
        return mime_type

    @skip_local_cache
    def download(self, from_path: str, to_path: str):
        """Downloads the file mentioned in from_path to to_path on the local
        system. The instance calling the method needs to be the FileStorage
        initialised with the remote file system.

        Args:
            from_path (str): Path of the file to be downloaded (remote)
            to_path (str): Path where the file is to be downloaded
            on local system

        Returns:
            NA
        """
        self.fs.get(rpath=from_path, lpath=to_path)

    @skip_local_cache
    def upload(self, from_path: str, to_path: str):
        """Uploads the file mentioned in from_path (local system) to to_path
        (remote system). The instance calling the method needs to be the
        FileStorage initialised with the remote file system where the file
        needs to be uploaded.

        Args:
            from_path (str): Path of the file to be uploaded (local)
            to_path (str): Path where the file is to be uploaded (usually remote)

        Returns:
            NA
        """
        self.fs.put(from_path, to_path)

    def glob(self, path: str) -> list[str]:
        """Lists files under path matching the pattern sepcified as part of
        path in the argument.

        Args:
            path (str): path to the directory where files matching the
            specified pattern is to be found
            Eg. a/b/c/*.txt will list all txt files under a/b/c/

        Returns:
            list[str]: List of file names matching any pattern specified
        """
        try:
            return self.fs.glob(path)
        except Exception as e:
            raise FileOperationError(str(e)) from e

    @skip_local_cache
    def get_hash_from_file(self, path: str) -> str:
        """Computes the hash for a file.

        Uses sha256 to compute the file hash through a buffered read.

        Args:
            file_path (str): Path to file that needs to be hashed

        Returns:
            str: SHA256 hash of the file
        """
        h = sha256()
        b = bytearray(128 * 1024)
        mv = memoryview(b)
        with self.fs.open(path) as f:
            while n := f.readinto(mv):
                h.update(mv[:n])
        return str(h.hexdigest())

    def json_dump(
        self,
        path: str,
        data: dict[str, Any],
        **kwargs: dict[Any, Any],  # type: ignore
    ):
        """Dumps data into the given file specified by path.

        Args:
            path (str): Path to file where JSON is to be dumped
            data (dict): Object to be written to the file
            **kwargs (dict): Any other additional arguments
        """
        try:
            with self.fs.open(path=path, mode="w", encoding="utf-8") as f:
                json.dump(obj=data, fp=f, **kwargs)  # type: ignore
        except Exception as e:
            raise FileOperationError(str(e)) from e

    def yaml_dump(
        self,
        path: str,
        data: dict[str, Any],
        **kwargs: dict[Any, Any],  # type: ignore
    ):
        """Dumps data into the given file specified by path.

        Args:
            path (str): Path to file where yml is to be dumped
            data (dict): Object to be written to the file
            **kwargs (dict): Any other additional arguments
        """
        try:
            with self.fs.open(path=path, mode="w", encoding="utf-8") as f:
                yaml.dump(data=data, stream=f, **kwargs)  # type: ignore
        except Exception as e:
            raise FileOperationError(str(e)) from e

    @skip_local_cache
    def json_load(self, path: str) -> dict[Any, Any]:
        with self.fs.open(path=path) as json_file:
            data: dict[str, Any] = json.load(json_file)
            return data

    @skip_local_cache
    def yaml_load(
        self,
        path: str,
    ) -> dict[Any, Any]:
        """Loads data from a file as yaml.

        Args:
            path (str): Path to file where yml is to be loaded

        Returns:
            dict[Any, Any]: Data loaded as yaml
        """
        with self.fs.open(path=path) as f:
            data: dict[str, Any] = yaml.safe_load(f)
            return data

    def guess_extension(self, path: str) -> str:
        """Returns the extension of the file passed.

        Args:
            path (str): String holding the path

        Returns:
            str: File extension
        """
        file_extension = ""
        sample_contents = self.read(
            path=path,
            mode="rb",
            length=FileOperationParams.EXTENSION_DEFAULT_READ_LENGTH,
        )
        if sample_contents:
            file_type = filetype.guess(sample_contents)
            file_extension = file_type.EXTENSION
        return file_extension

    def walk(self, path: str, max_depth=None, topdown=True):
        """Walks the dir in the path and returns the list of files/dirs.

        Args:
            path (str): Root to recurse into
            maxdepth (int): Maximum recursion depth. None means limitless,
            but not recommended
                on link-based file-systems.
            topdown (bool): Whether to walk the directory tree from the top
            downwards or from
                the bottom upwards.

        Returns:
            Iterator containing the list of files and folders
        """
        # Invalidating cache explicitly to avoid any stale listing
        self.fs.invalidate_cache(path=path)
        return self.fs.walk(path, maxdepth=max_depth, topdown=topdown)



import json
import logging
from io import BytesIO
from pathlib import Path
from typing import Any

import requests
from requests import Response
from requests.exceptions import ConnectionError, HTTPError, Timeout
from unstract.llmwhisperer.client_v2 import (
    LLMWhispererClientException,
    LLMWhispererClientV2,
)
from unstract.sdk.adapters.exceptions import ExtractorError
from unstract.sdk.adapters.utils import AdapterUtils
from unstract.sdk.adapters.x2text.constants import X2TextConstants
from unstract.sdk.adapters.x2text.llm_whisperer_v2.src.constants import (
    Modes,
    OutputModes,
    WhispererConfig,
    WhispererDefaults,
    WhispererHeader,
    WhisperStatus,
)
from unstract.sdk.adapters.x2text.llm_whisperer_v2.src.dto import WhispererRequestParams
from unstract.sdk.constants import MimeType
from unstract.sdk.file_storage import FileStorage, FileStorageProvider

logger = logging.getLogger(__name__)


class LLMWhispererHelper:
    @staticmethod
    def get_request_headers(config: dict[str, Any]) -> dict[str, Any]:
        """Obtains the request headers to authenticate with LLMWhisperer.

        Returns:
            str: Request headers
        """
        return {
            "accept": MimeType.JSON,
            WhispererHeader.UNSTRACT_KEY: config.get(WhispererConfig.UNSTRACT_KEY),
        }

    @staticmethod
    def test_connection_request(
        config: dict[str, Any], request_endpoint: str
    ) -> Response:
        llm_whisperer_svc_url = f"{config.get(WhispererConfig.URL)}" f"/api/v2"
        headers = LLMWhispererHelper.get_request_headers(config=config)

        try:
            response: Response
            url = f"{llm_whisperer_svc_url}/{request_endpoint}"
            response = requests.get(url=url, headers=headers)
            response.raise_for_status()
        except ConnectionError as e:
            logger.error(f"Adapter error: {e}")
            raise ExtractorError(
                "Unable to connect to LLMWhisperer service, please check the URL",
                actual_err=e,
                status_code=503,
            ) from e
        except Timeout as e:
            msg = "Request to LLMWhisperer has timed out"
            logger.error(f"{msg}: {e}")
            raise ExtractorError(msg, actual_err=e, status_code=504) from e
        except HTTPError as e:
            logger.error(f"Adapter error: {e}")
            default_err = "Error while calling the LLMWhisperer service"
            msg = AdapterUtils.get_msg_from_request_exc(
                err=e, message_key="message", default_err=default_err
            )
            raise ExtractorError(
                msg, status_code=e.response.status_code, actual_err=e
            ) from e

    @staticmethod
    def make_request(
        config: dict[str, Any],
        headers: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
        data: BytesIO | None = None,
        type: str = "whisper",
    ) -> Response:
        """Makes a request to LLMWhisperer service.

        Args:
            config (dict[str, Any]): LLMWhisperer config to use
            headers (Optional[dict[str, Any]], optional): Headers to pass.
                Defaults to None.
            params (Optional[dict[str, Any]], optional): Query params to pass.
                Defaults to None.
            data (Optional[BytesIO], optional): Data to pass in case of POST.
                Defaults to None.
            type (str, optional): Type of request / endpoint in LLMWhisperer.
                Defaults to "whisper".

        Returns:
            Response: Response from the request
        """
        llm_whisperer_svc_url = f"{config.get(WhispererConfig.URL)}" f"/api/v2"
        if not headers:
            headers = LLMWhispererHelper.get_request_headers(config=config)

        try:
            response: dict[str, Any]
            client = LLMWhispererClientV2(
                base_url=llm_whisperer_svc_url,
                api_key=config.get(WhispererConfig.UNSTRACT_KEY),
                logging_level=WhispererDefaults.LOGGING_LEVEL,
            )
            if type == "whisper":
                response = client.whisper(**params, stream=data)
                if response["status_code"] == 200:
                    logger.debug(
                        "Successfully extracted for whisper hash: "
                        f"{response.get(X2TextConstants.WHISPER_HASH_V2, '')}"
                    )
                    response["extraction"][X2TextConstants.WHISPER_HASH_V2] = (
                        response.get(X2TextConstants.WHISPER_HASH_V2, "")
                    )
                    return response["extraction"]
                else:
                    response["message"] += (
                        ". Whisper hash: "
                        f"{response.get(X2TextConstants.WHISPER_HASH_V2, '')}"
                    )
                    raise ExtractorError(
                        response["message"],
                        response["status_code"],
                        actual_err=response,
                    )
            elif type == "highlight":
                response = client.get_highlight_data(**params)
                return response

        except ConnectionError as e:
            logger.error(f"Adapter error: {e}")
            raise ExtractorError(
                "Unable to connect to LLMWhisperer service, please check the URL",
                actual_err=e,
                status_code=503,
            ) from e
        except Timeout as e:
            msg = "Request to LLMWhisperer has timed out"
            logger.error(f"{msg}: {e}")
            raise ExtractorError(msg, actual_err=e, status_code=504) from e
        except LLMWhispererClientException as e:
            logger.error(f"LLM Whisperer error: {e}")
            raise ExtractorError(
                message=f"LLM Whisperer error: {e}",
                actual_err=e,
                status_code=500,
            ) from e

        return response

    @staticmethod
    def get_whisperer_params(
        config: dict[str, Any], extra_params: WhispererRequestParams
    ) -> dict[str, Any]:
        """Gets query params meant for /whisper endpoint.

        The params is filled based on the configuration passed.

        Returns:
            dict[str, Any]: Query params
        """
        params = {
            WhispererConfig.MODE: config.get(WhispererConfig.MODE, Modes.FORM.value),
            WhispererConfig.OUTPUT_MODE: config.get(
                WhispererConfig.OUTPUT_MODE, OutputModes.LAYOUT_PRESERVING.value
            ),
            WhispererConfig.LINE_SPLITTER_TOLERANCE: config.get(
                WhispererConfig.LINE_SPLITTER_TOLERANCE,
                WhispererDefaults.LINE_SPLITTER_TOLERANCE,
            ),
            WhispererConfig.LINE_SPLITTER_STRATEGY: config.get(
                WhispererConfig.LINE_SPLITTER_STRATEGY,
                WhispererDefaults.LINE_SPLITTER_STRATEGY,
            ),
            WhispererConfig.HORIZONTAL_STRETCH_FACTOR: config.get(
                WhispererConfig.HORIZONTAL_STRETCH_FACTOR,
                WhispererDefaults.HORIZONTAL_STRETCH_FACTOR,
            ),
            WhispererConfig.PAGES_TO_EXTRACT: config.get(
                WhispererConfig.PAGES_TO_EXTRACT,
                WhispererDefaults.PAGES_TO_EXTRACT,
            ),
            WhispererConfig.MARK_VERTICAL_LINES: config.get(
                WhispererConfig.MARK_VERTICAL_LINES,
                WhispererDefaults.MARK_VERTICAL_LINES,
            ),
            WhispererConfig.MARK_HORIZONTAL_LINES: config.get(
                WhispererConfig.MARK_HORIZONTAL_LINES,
                WhispererDefaults.MARK_HORIZONTAL_LINES,
            ),
            WhispererConfig.PAGE_SEPARATOR: config.get(
                WhispererConfig.PAGE_SEPARATOR,
                WhispererDefaults.PAGE_SEPARATOR,
            ),
            WhispererConfig.ADD_LINE_NOS: extra_params.enable_highlight,
            # Not providing default value to maintain legacy compatablity
            # these are optional params and identifiers for audit
            WhispererConfig.TAG: extra_params.tag
            or config.get(
                WhispererConfig.TAG,
                WhispererDefaults.TAG,
            ),
            WhispererConfig.USE_WEBHOOK: config.get(WhispererConfig.USE_WEBHOOK, ""),
            WhispererConfig.WEBHOOK_METADATA: config.get(
                WhispererConfig.WEBHOOK_METADATA
            ),
            WhispererConfig.WAIT_TIMEOUT: config.get(
                WhispererConfig.WAIT_TIMEOUT,
                WhispererDefaults.WAIT_TIMEOUT,
            ),
            WhispererConfig.WAIT_FOR_COMPLETION: WhispererDefaults.WAIT_FOR_COMPLETION,
        }
        if params[WhispererConfig.MODE] == Modes.LOW_COST.value:
            params.update(
                {
                    WhispererConfig.MEDIAN_FILTER_SIZE: config.get(
                        WhispererConfig.MEDIAN_FILTER_SIZE,
                        WhispererDefaults.MEDIAN_FILTER_SIZE,
                    ),
                    WhispererConfig.GAUSSIAN_BLUR_RADIUS: config.get(
                        WhispererConfig.GAUSSIAN_BLUR_RADIUS,
                        WhispererDefaults.GAUSSIAN_BLUR_RADIUS,
                    ),
                }
            )
        return params

    @staticmethod
    def send_whisper_request(
        input_file_path: str,
        config: dict[str, Any],
        extra_params: WhispererRequestParams,
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> requests.Response:
        params = LLMWhispererHelper.get_whisperer_params(
            config=config, extra_params=extra_params
        )
        response: requests.Response
        try:
            input_file_data = BytesIO(fs.read(path=input_file_path, mode="rb"))
            enable_highlight = extra_params.enable_highlight
            response = LLMWhispererHelper.make_request(
                config=config,
                params=params,
                data=input_file_data,
            )
            if enable_highlight:
                whisper_hash = response.get(X2TextConstants.WHISPER_HASH_V2, "")
                highlight_data = LLMWhispererHelper.make_highlight_data_request(
                    config,
                    whisper_hash,
                    enable_highlight,
                )
                response["line_metadata"] = highlight_data
        except OSError as e:
            logger.error(f"OS error while reading {input_file_path}: {e}")
            raise ExtractorError(str(e)) from e
        return response

    @staticmethod
    def make_highlight_data_request(
        config: dict[str, Any], whisper_hash: str, enable_highlight: bool
    ) -> dict[Any, Any]:
        """Makes a call to get highlight data from LLMWhisperer.

        Args:
            config (dict[str, Any]): LLMWhisperer config to use
            whisper_hash (str): Identifier of the extraction
            enable_highlight (bool): Whether to enable highlight

        Returns:
            dict[Any, Any]: Highlight data
        """
        logger.info(f"Extracting async for whisper hash: {whisper_hash}")

        headers: dict[str, Any] = LLMWhispererHelper.get_request_headers(config)
        params = {
            WhisperStatus.WHISPER_HASH: whisper_hash,
            WhispererConfig.EXTRACT_ALL_LINES: enable_highlight,
            WhispererConfig.LINES: "",
        }

        retrieve_response = LLMWhispererHelper.make_request(
            config=config,
            headers=headers,
            params=params,
            type="highlight",
        )
        return retrieve_response

    @staticmethod
    def extract_text_from_response(
        output_file_path: str | None,
        response: dict[str, Any],
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> str:
        if not response:
            raise ExtractorError("Couldn't extract text from file", status_code=500)
        output_json = {}
        output_json = response
        if output_file_path:
            LLMWhispererHelper.write_output_to_file(
                output_json=output_json,
                output_file_path=Path(output_file_path),
                fs=fs,
            )
        return output_json.get("result_text", "")

    @staticmethod
    def write_output_to_file(
        output_json: dict,
        output_file_path: Path,
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> None:
        """Write LLMW outputs to file.

        Writes the extracted text and metadata to the specified output file
        and metadata file.

        Args:
            output_json (dict): The dictionary containing the extracted data,
                with "text" as the key for the main content.
            output_file_path (Path): The file path where the extracted text
                should be written.
            fs (FileStorage): File storage instance to use for writing

        Raises:
            ExtractorError: If there is an error while writing the output file.
        """
        try:
            text_output = output_json.get("result_text", "")
            logger.info(f"Writing output to {output_file_path}")
            fs.write(
                path=str(output_file_path),
                mode="w",
                data=text_output,
                encoding="utf-8",
            )
        except Exception as e:
            logger.error(f"Error while writing {output_file_path}: {e}")
            raise ExtractorError(str(e)) from e
        try:
            # Define the directory of the output file and metadata paths
            output_dir = output_file_path.parent
            metadata_dir = output_dir / "metadata"
            metadata_file_name = output_file_path.with_suffix(".json").name
            metadata_file_path = metadata_dir / metadata_file_name
            # Ensure the metadata directory exists
            fs.mkdir(create_parents=True, path=str(metadata_dir))
            # Remove the "result_text" key from the metadata
            metadata = {
                key: value for key, value in output_json.items() if key != "result_text"
            }
            metadata_json = json.dumps(metadata, ensure_ascii=False, indent=4)
            logger.info(f"Writing metadata to {metadata_file_path}")
            fs.write(
                path=str(metadata_file_path),
                mode="w",
                data=metadata_json,
                encoding="utf-8",
            )
        except Exception as e:
            logger.warn(f"Error while writing metadata to {metadata_file_path}: {e}")



import logging
from typing import Any

from unstract.sdk.file_storage import FileStorage, FileStorageProvider
from unstract.sdk.utils import ToolUtils

logger = logging.getLogger(__name__)


class ToolConfigHelper:
    """Helper class to handle static commands for tools."""

    @staticmethod
    def spec(
        spec_file: str = "config/spec.json",
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> dict[str, Any]:
        """Returns the JSON schema for the tool settings.

        Args:
            spec_file (str): The path to the JSON schema file.
            The default is config/spec.json.

        Returns:
            str: The JSON schema of the tool.
        """
        return ToolUtils.load_json(spec_file, fs=fs)

    @staticmethod
    def properties(
        properties_file: str = "config/properties.json",
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> dict[str, Any]:
        """Returns the properties of the tool.

        Args:
            properties_file (str): The path to the properties file.
            The default is config/properties.json.

        Returns:
            str: The properties of the tool.
        """
        return ToolUtils.load_json(properties_file, fs)

    @staticmethod
    def variables(
        variables_file: str = "config/runtime_variables.json",
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> dict[str, Any]:
        """Returns the JSON schema of the runtime variables.

        Args:
            variables_file (str): The path to the JSON schema file.
            The default is config/runtime_variables.json.

        Returns:
            str: The JSON schema for the runtime variables.
        """
        try:
            return ToolUtils.load_json(variables_file, fs)
        # Allow runtime variables definition to be optional
        except FileNotFoundError:
            logger.info("No runtime variables defined for tool")
            return {}

    @staticmethod
    def icon(
        icon_file: str = "config/icon.svg",
        fs: FileStorage = FileStorage(provider=FileStorageProvider.LOCAL),
    ) -> str:
        """Returns the icon of the tool.

        Args:
            icon_file (str): The path to the icon file.
                The default is config/icon.svg.

        Returns:
            str: The icon of the tool.
        """
        icon = fs.read(path=icon_file, mode="rb", encoding="utf-8")
        return icon
